#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   eval_classifier 
@Time        :   2023/9/22 15:22
@Author      :   Xuesong Chen
@Description :   
"""

from wuji.Evaluator.Classification import ClassificationEvaluator
if __name__ == '__main__':
    '''
        睡眠分期的线上版本使用accuracy, kappa和confusion_matrix
    '''
    gt_labels = ['Wake', 'N1', 'N2', 'N3', 'REM', 'N2', 'N3', 'REM']        # 用户标注结果
    pred_labels = ['Wake', 'N1', 'N2', 'REM', 'REM', 'N3', 'N3', 'REM']     # 算法预测结果
    evaluator = ClassificationEvaluator(gt_labels, pred_labels)
    print(evaluator.accuracy())
    print(evaluator.precision())
    print(evaluator.recall())
    print(evaluator.f1())
    print(evaluator.confusion_matrix(normalize=None))
    print(evaluator.kappa())
    print(evaluator.confusion_matrix(normalize='true', labels=['Wake', 'N1', 'N2', 'N3', 'REM']))




#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   signal 
@Time        :   2023/11/21 14:06
@Author      :   Xuesong Chen
@Description :   
"""
import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import NDArray
from hrvanalysis import get_nn_intervals
import neurokit2 as nk
from scipy.signal import filtfilt
import typing as t

def interp_heartbeat(positions, values, fs, target_fs, n_samples):
    '''
    Interpolate RRI or JJI to target sampling frequency
    Parameters
    ----------
    positions: positions of RRI or JJI
    values: values of RRI or JJI
    fs: sampling frequency of RRI or JJI
    target_fs: target sampling frequency
    n_samples: number of samples of the raw signal

    Returns
    -------
    interp_value: interpolated RRI or JJI at target sampling frequency, in ms
    '''
    rri_in_ms = values / fs * 1000
    filtered_rri_in_ms = get_nn_intervals(rri_in_ms, verbose=False, ectopic_beats_removal_method='karlsson')
    interp_value = np.interp(
        np.arange(0, n_samples, fs / target_fs), positions,
        filtered_rri_in_ms).astype(np.float32)
    interp_value[np.isnan(interp_value)] = np.nanmean(interp_value)
    return interp_value


def filter_spo2(spo2, fs, min_allowed_spo2=50, max_allowed_spo2=105):
    """
    This function filters the SpO2 signal based on certain conditions.

    Parameters:
    spo2 (numpy array): The SpO2 signal data.
    fs (int): The sampling frequency of the SpO2 signal.
    min_allowed_spo2 (int, optional): The minimum allowed SpO2 value. Defaults to 50.
    max_allowed_spo2 (int, optional): The maximum allowed SpO2 value. Defaults to 105.

    Returns:
    spo2 (numpy array): The filtered SpO2 signal data.
    """
    if fs != 1:
        signal = nk.signal_resample(signal, sampling_rate=sfreq, desired_sampling_rate=1)
    plt.plot(spo2, label='raw')
    # 如果下降速率大于5，或者上升速率大于10，则认为是异常值
    max_drop_rate = -5
    max_recover_rate = 10
    mask = np.logical_or(np.diff(spo2) < max_drop_rate, np.diff(spo2) > max_recover_rate)
    spo2[1:][mask] = np.nan
    mask = np.logical_or(spo2 < min_allowed_spo2, spo2 > max_allowed_spo2)
    spo2[mask] = np.nan
    return spo2


def get_eeg_features(eeg: NDArray, fs: t.Union[int, float]) -> NDArray:
    """
    This function calculates the normalized power of an EEG signal.

    Parameters:
    eeg (numpy array): The EEG signal data.
    fs (int): The sampling frequency of the EEG signal.

    Returns:
    norm_power (numpy array): The normalized power of the EEG signal. The dimensions are frequency by time.
    """
    seg_duration = 1
    freq, time, power = nk.signal_timefrequency(
        eeg,
        window=seg_duration,
        overlap=fs * seg_duration * 0.5,
        min_frequency=0.3,
        max_frequency=35,
        sampling_rate=fs,
        show=False)
    norm_power = power / (np.sum(power, axis=0, keepdims=True) + 1e-8)
    return norm_power  # freq * time



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
I/O
==================
Functions for loading annotations from files in various formats.

.. autosummary::
    :toctree: generated/

    load_event_list
    load_scene_list
    load_file_pair_list

"""

from __future__ import absolute_import
import csv
import dcase_util


def load_event_list(filename, **kwargs):
    """Load event list from csv formatted text-file

    Supported formats (see more `dcase_util.containers.MetaDataContainer.load()` method):

    - [event onset (float >= 0)][delimiter][event offset (float >= 0)]
    - [event onset (float >= 0)][delimiter][event offset (float >= 0)][delimiter][label]
    - [filename][delimiter][event onset (float >= 0)][delimiter][event offset (float >= 0)][delimiter][event label]
    - [filename][delimiter][scene_label][delimiter][event onset (float >= 0)][delimiter][event offset (float >= 0)][delimiter][event label]
    - [filename]

    Supported delimiters: ``,``, ``;``, ``tab``

    Example of event list file::

        21.64715	23.00552	alert
        36.91184	38.27021	alert
        69.72575	71.09029	alert
        63.53990	64.89827	alert
        84.25553	84.83920	alert
        20.92974	21.82661	clearthroat
        28.39992	29.29679	clearthroat
        80.47837	81.95937	clearthroat
        44.48363	45.96463	clearthroat
        78.13073	79.05953	clearthroat
        15.17031	16.27235	cough
        20.54931	21.65135	cough
        27.79964	28.90168	cough
        75.45959	76.32490	cough
        70.81708	71.91912	cough
        21.23203	22.55902	doorslam
        7.546220	9.014880	doorslam
        34.11303	35.04183	doorslam
        45.86001	47.32867	doorslam


    Parameters
    ----------
    filename : str
        Path to the csv-file

    Returns
    -------
    list of dict
        Event list

    """

    return dcase_util.containers.MetaDataContainer().load(filename=filename, **kwargs)


def load_scene_list(filename, **kwargs):
    """Load scene list from csv formatted text-file

    Supported formats (see more `dcase_util.containers.MetaDataContainer.load()` method):

    - [filename][delimiter][scene label]
    - [filename][delimiter][segment start (float >= 0)][delimiter][segment stop (float >= 0)][delimiter][scene label]

    Supported delimiters: ``,``, ``;``, ``tab``

    Example of scene list file::

        scenes_stereo/supermarket09.wav	supermarket
        scenes_stereo/tubestation10.wav	tubestation
        scenes_stereo/quietstreet08.wav	quietstreet
        scenes_stereo/restaurant05.wav	restaurant
        scenes_stereo/busystreet05.wav	busystreet
        scenes_stereo/openairmarket04.wav	openairmarket
        scenes_stereo/quietstreet01.wav	quietstreet
        scenes_stereo/supermarket05.wav	supermarket
        scenes_stereo/openairmarket01.wav	openairmarket

    Parameters
    ----------
    filename : str
        Path to the csv-file

    Returns
    -------
    list of dict
        Scene list

    """

    return dcase_util.containers.MetaDataContainer().load(filename=filename, **kwargs)


def load_file_pair_list(filename):
    """Load file pair list csv formatted text-file

    Format is [reference_file][delimiter][estimated_file]

    Supported delimiters: ``,``, ``;``, ``tab``

    Example of file-list::

        office_snr0_high_v2.txt	office_snr0_high_v2_detected.txt
        office_snr0_med_v2.txt	office_snr0_med_v2_detected.txt

    Parameters
    ----------
    filename : str
        Path to the csv-file

    Returns
    -------
    file_list: list
        File pair dicts in a list

    """

    data = []

    input_file = open(filename, 'rt')

    try:
        dialect = csv.Sniffer().sniff(input_file.readline(), [',', ';', '\t'])

    except csv.Error:
        raise ValueError('Unknown delimiter in file [{file}].'.format(file=filename))

    input_file.seek(0)  

    for row in csv.reader(input_file, dialect):
        if len(row) == 2:
            data.append(
                {
                    'reference_file': row[0],
                    'estimated_file': row[1]
                }
            )

        else:
            raise ValueError('Unknown file pair list format in file [{file}].'.format(file=filename))

    input_file.close()

    return data




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Segment-based metrics, main functions:

* :func:`sed_eval.sound_event.SegmentBasedMetrics.evaluate`: Calculate intermediate values for evaluation and accumulate them.
* :func:`sed_eval.sound_event.SegmentBasedMetrics.results`: Calculate and return all metrics.
* :func:`sed_eval.sound_event.SegmentBasedMetrics.results_overall_metrics`: Calculate and return overall metrics (micro-averaged).
* :func:`sed_eval.sound_event.SegmentBasedMetrics.results_class_wise_metrics`: Calculate and return class-wise metrics.
* :func:`sed_eval.sound_event.SegmentBasedMetrics.results_class_wise_average_metrics`: Calculate and return class-wise average metrics (macro-averaged).

Event-based metrics, main functions:

* :func:`sed_eval.sound_event.EventBasedMetrics.evaluate`: Calculate intermediate values for evaluation and accumulate them.
* :func:`sed_eval.sound_event.EventBasedMetrics.results`: Calculate and return all metrics.
* :func:`sed_eval.sound_event.EventBasedMetrics.results_overall_metrics`: Calculate and return overall metrics (micro-averaged).
* :func:`sed_eval.sound_event.EventBasedMetrics.results_class_wise_metrics`: Calculate and return class-wise metrics.
* :func:`sed_eval.sound_event.EventBasedMetrics.results_class_wise_average_metrics`: Calculate and return class-wise average metrics (macro-averaged).

Functions :func:`sed_eval.sound_event.SegmentBasedMetrics.evaluate` and :func:`sed_eval.sound_event.EventBasedMetrics.evaluate`
take as a parameter event lists, use :func:`sed_eval.io.load_event_list` to read them from a file.


Usage example when reading event lists from disk (you can run example in path ``tests/data/sound_event``):

.. code-block:: python
    :linenos:

    import sed_eval
    import dcase_util

    file_list = [
        {
         'reference_file': 'office_snr0_high_v2.txt',
         'estimated_file': 'office_snr0_high_v2_detected.txt'
        },
        {
         'reference_file': 'office_snr0_med_v2.txt',
         'estimated_file': 'office_snr0_med_v2_detected.txt'
        }
    ]

    data = []

    # Get used event labels
    all_data = dcase_util.containers.MetaDataContainer()
    for file_pair in file_list:
        reference_event_list = sed_eval.io.load_event_list(
            filename=file_pair['reference_file']
        )
        estimated_event_list = sed_eval.io.load_event_list(
            filename=file_pair['estimated_file']
        )

        data.append({'reference_event_list': reference_event_list,
                     'estimated_event_list': estimated_event_list})

        all_data += reference_event_list

    event_labels = all_data.unique_event_labels

    # Start evaluating

    # Create metrics classes, define parameters
    segment_based_metrics = sed_eval.sound_event.SegmentBasedMetrics(
        event_label_list=event_labels,
        time_resolution=1.0
    )

    event_based_metrics = sed_eval.sound_event.EventBasedMetrics(
        event_label_list=event_labels,
        t_collar=0.250
    )

    # Go through files
    for file_pair in data:
        segment_based_metrics.evaluate(
            reference_event_list=file_pair['reference_event_list'],
            estimated_event_list=file_pair['estimated_event_list']
        )

        event_based_metrics.evaluate(
            reference_event_list=file_pair['reference_event_list'],
            estimated_event_list=file_pair['estimated_event_list']
        )

    # Get only certain metrics
    overall_segment_based_metrics = segment_based_metrics.results_overall_metrics()
    print("Accuracy:", overall_segment_based_metrics['accuracy']['accuracy'])

    # Or print all metrics as reports
    print(segment_based_metrics)
    print(event_based_metrics)

Usage example to evaluate results stored in variables:

.. code-block:: python
    :linenos:

    import sed_eval
    import dcase_util

    reference_event_list = dcase_util.containers.MetaDataContainer(
        [
            {
                'event_label': 'car',
                'event_onset': 0.0,
                'event_offset': 2.5,
                'file': 'audio/street/b099.wav',
                'scene_label': 'street'
            },
            {
                'event_label': 'car',
                'event_onset': 2.8,
                'event_offset': 4.5,
                'file': 'audio/street/b099.wav',
                'scene_label': 'street'
            },
            {
                'event_label': 'car',
                'event_onset': 6.0,
                'event_offset': 10.0,
                'file': 'audio/street/b099.wav',
                'scene_label': 'street'
            }
        ]
    )

    estimated_event_list = dcase_util.containers.MetaDataContainer(
        [
            {
                'event_label': 'car',
                'event_onset': 1.0,
                'event_offset': 3.5,
                'file': 'audio/street/b099.wav',
                'scene_label': 'street'
            },
            {
                'event_label': 'car',
                'event_onset': 7.0,
                'event_offset': 8.0,
                'file': 'audio/street/b099.wav',
                'scene_label': 'street'
            }
        ]
    )

    segment_based_metrics = sed_eval.sound_event.SegmentBasedMetrics(
        event_label_list=reference_event_list.unique_event_labels,
        time_resolution=1.0
    )
    event_based_metrics = sed_eval.sound_event.EventBasedMetrics(
        event_label_list=reference_event_list.unique_event_labels,
        t_collar=0.250
    )

    for filename in reference_event_list.unique_files:
        reference_event_list_for_current_file = reference_event_list.filter(
            filename=filename
        )

        estimated_event_list_for_current_file = estimated_event_list.filter(
            filename=filename
        )

        segment_based_metrics.evaluate(
            reference_event_list=reference_event_list_for_current_file,
            estimated_event_list=estimated_event_list_for_current_file
        )

        event_based_metrics.evaluate(
            reference_event_list=reference_event_list_for_current_file,
            estimated_event_list=estimated_event_list_for_current_file
        )

    # Get only certain metrics
    overall_segment_based_metrics = segment_based_metrics.results_overall_metrics()
    print("Accuracy:", overall_segment_based_metrics['accuracy']['accuracy'])

    # Or print all metrics as reports
    print(segment_based_metrics)
    print(event_based_metrics)

Segment based metrics
^^^^^^^^^^^^^^^^^^^^^

.. autosummary::
    :toctree: generated/

    SegmentBasedMetrics
    SegmentBasedMetrics.evaluate
    SegmentBasedMetrics.results
    SegmentBasedMetrics.results_overall_metrics
    SegmentBasedMetrics.results_class_wise_metrics
    SegmentBasedMetrics.results_class_wise_average_metrics
    SegmentBasedMetrics.result_report_parameters
    SegmentBasedMetrics.result_report_overall
    SegmentBasedMetrics.result_report_class_wise
    SegmentBasedMetrics.result_report_class_wise_average
    SegmentBasedMetrics.reset

.. autoclass:: SegmentBasedMetrics
   :members:

Event based metrics
^^^^^^^^^^^^^^^^^^^

.. autosummary::
    :toctree: generated/

    EventBasedMetrics
    EventBasedMetrics.evaluate
    EventBasedMetrics.results
    EventBasedMetrics.results_overall_metrics
    EventBasedMetrics.results_class_wise_metrics
    EventBasedMetrics.results_class_wise_average_metrics
    EventBasedMetrics.result_report_parameters
    EventBasedMetrics.result_report_overall
    EventBasedMetrics.result_report_class_wise
    EventBasedMetrics.result_report_class_wise_average
    EventBasedMetrics.reset

.. autoclass:: EventBasedMetrics
   :members:
   :undoc-members:
   :inherited-members:

"""

from __future__ import absolute_import
import numpy
import math
import dcase_util
from . import metric
from . import util


class SoundEventMetrics(object):
    """Base class for sound event detection metrics.

    """

    def __init__(self,
                 empty_system_output_handling=None):
        """Constructor

        Parameters
        ----------

        empty_system_output_handling : str
            Controls how empty system output is handled, i.e. when Nsys = 0. Default behaviour is to show NaN when e.g.
            computing precision (Ntp / Nsys).
            Use 'zero_score' to force these score to zero.
            Default value None

        """

        self.event_label_list = []
        self.ui = dcase_util.ui.FancyStringifier()
        self.empty_system_output_handling = empty_system_output_handling

    # Reports
    def result_report_overall(self):
        """Report overall results

        Returns
        -------
        str
            result report in string format

        """

        results = self.results_overall_metrics()

        output = self.ui.section_header('Overall metrics (micro-average)', indent=2) + '\n'

        if results['f_measure']:
            output += self.ui.line('F-measure', indent=2) + '\n'
            output += self.ui.data(field='F-measure (F1)', value=float(results['f_measure']['f_measure']) * 100,
                                   unit='%', indent=4) + '\n'
            output += self.ui.data(field='Precision', value=float(results['f_measure']['precision']) * 100,
                                   unit='%', indent=4) + '\n'
            output += self.ui.data(field='Recall', value=float(results['f_measure']['recall']) * 100,
                                   unit='%', indent=4) + '\n'

        if results['error_rate']:
            output += self.ui.line('Error rate', indent=2) + '\n'
            output += self.ui.data(field='Error rate (ER)', value=float(results['error_rate']['error_rate']),
                                   indent=4) + '\n'
            output += self.ui.data(field='Substitution rate', value=float(results['error_rate']['substitution_rate']),
                                   indent=4) + '\n'
            output += self.ui.data(field='Deletion rate', value=float(results['error_rate']['deletion_rate']),
                                   indent=4) + '\n'
            output += self.ui.data(field='Insertion rate', value=float(results['error_rate']['insertion_rate']),
                                   indent=4) + '\n'

        if results['accuracy']:
            output += self.ui.line('Accuracy', indent=2) + '\n'
            output += self.ui.data(field='Sensitivity', value=float(results['accuracy']['sensitivity'] * 100),
                                   unit='%', indent=4) + '\n'
            output += self.ui.data(field='Specificity', value=float(results['accuracy']['specificity'] * 100),
                                   unit='%', indent=4) + '\n'
            output += self.ui.data(field='Balanced accuracy',
                                   value=float(results['accuracy']['balanced_accuracy'] * 100),
                                   unit='%', indent=4) + '\n'
            output += self.ui.data(field='Accuracy', value=float(results['accuracy']['accuracy'] * 100),
                                   unit='%', indent=4) + '\n'

        return output

    def result_report_class_wise_average(self):
        """Report class-wise averages

        Returns
        -------
        str
            result report in string format

        """

        results = self.results_class_wise_average_metrics()

        output = self.ui.section_header('Class-wise average metrics (macro-average)', indent=2) + '\n'

        if results['f_measure']:
            output += self.ui.line('F-measure', indent=2) + '\n'
            output += self.ui.data(field='F-measure (F1)', value=float(results['f_measure']['f_measure']) * 100,
                                   unit='%', indent=4) + '\n'
            output += self.ui.data(field='Precision', value=float(results['f_measure']['precision']) * 100,
                                   unit='%', indent=4) + '\n'
            output += self.ui.data(field='Recall', value=float(results['f_measure']['recall']) * 100,
                                   unit='%', indent=4) + '\n'

        if results['error_rate']:
            output += self.ui.line('Error rate', indent=2) + '\n'
            output += self.ui.data(field='Error rate (ER)', value=float(results['error_rate']['error_rate']),
                                   indent=4) + '\n'
            output += self.ui.data(field='Deletion rate', value=float(results['error_rate']['deletion_rate']),
                                   indent=4) + '\n'
            output += self.ui.data(field='Insertion rate', value=float(results['error_rate']['insertion_rate']),
                                   indent=4) + '\n'

        if results['accuracy']:
            output += self.ui.line('Accuracy', indent=2) + '\n'
            output += self.ui.data(field='Sensitivity', value=float(results['accuracy']['sensitivity'] * 100),
                                   unit='%', indent=4) + '\n'
            output += self.ui.data(field='Specificity', value=float(results['accuracy']['specificity'] * 100),
                                   unit='%', indent=4) + '\n'
            output += self.ui.data(field='Balanced accuracy',
                                   value=float(results['accuracy']['balanced_accuracy'] * 100),
                                   unit='%', indent=4) + '\n'
            output += self.ui.data(field='Accuracy', value=float(results['accuracy']['accuracy'] * 100),
                                   unit='%', indent=4) + '\n'

        output += "  \n"

        return output

    def result_report_class_wise(self):
        """Report class-wise results

        Returns
        -------
        str
            result report in string format

        """

        results = self.results_class_wise_metrics()

        accuracy_present = True
        for event_label in self.event_label_list:
            if 'accuracy' not in results[event_label]['accuracy']:
                accuracy_present = False

        output = self.ui.section_header('Class-wise metrics', indent=2) + '\n'

        headers = ['Event label', 'Nref', 'Nsys', 'F', 'Pre', 'Rec', 'ER', 'Del', 'Ins']
        sep = ['-', '-', '-', '-', '-', '-', '-', '-', '-']
        widths = [15, 8, 8, 9, 9, 9, 9, 9, 9]
        separators = [True, False, True, False, False, True, False, False, True]
        if accuracy_present:
            headers += ['Sens', 'Spec', 'Bacc', 'Acc']
            sep += ['-', '-', '-', '-']
            widths += [9, 9, 9, 9]
            separators += [False, False, False, False]

        output += self.ui.row(*headers, widths=widths, indent=4, separators=separators) + '\n'
        output += self.ui.row(*sep) + '\n'
        for event_label in self.event_label_list:
            data = [
                event_label,
                results[event_label]['count']['Nref'],
                results[event_label]['count']['Nsys'],
                results[event_label]['f_measure']['f_measure'] * 100,
                results[event_label]['f_measure']['precision'] * 100,
                results[event_label]['f_measure']['recall'] * 100,
                results[event_label]['error_rate']['error_rate'],
                results[event_label]['error_rate']['deletion_rate'],
                results[event_label]['error_rate']['insertion_rate']
            ]

            types = [
                'str15',
                'int',
                'int',
                'float1_percentage',
                'float1_percentage',
                'float1_percentage',
                'float2',
                'float2',
                'float2',
            ]

            if accuracy_present:
                data += [
                    results[event_label]['accuracy']['sensitivity'] * 100,
                    results[event_label]['accuracy']['specificity'] * 100,
                    results[event_label]['accuracy']['balanced_accuracy'] * 100,
                    results[event_label]['accuracy']['accuracy'] * 100
                ]

                types += [
                    'float1_percentage',
                    'float1_percentage',
                    'float1_percentage',
                    'float1_percentage',
                ]

            output += self.ui.row(*data, types=types) + '\n'

        return output

    # Metrics / overall
    def overall_f_measure(self):
        return {}

    def overall_error_rate(self):
        return {}

    def overall_accuracy(self, factor=0.5):
        return {}

    # Metrics / class-wise
    def class_wise_count(self, event_label):
        return {}

    def class_wise_f_measure(self, event_label):
        return {}

    def class_wise_error_rate(self, event_label):
        return {}

    def class_wise_accuracy(self, event_label):
        return {}

    # Results
    def results_overall_metrics(self):
        """Overall metrics

        Returns
        -------
        dict
            results in a dictionary format

        """

        return {
            'f_measure': self.overall_f_measure(),
            'error_rate': self.overall_error_rate(),
            'accuracy': self.overall_accuracy()
        }

    def results_class_wise_metrics(self):
        """Class-wise metrics

        Returns
        -------
        dict
            results in a dictionary format

        """

        results = {}
        for event_label in self.event_label_list:
            if event_label not in results:
                results[event_label] = {}

            results[event_label]['f_measure'] = self.class_wise_f_measure(event_label)
            results[event_label]['accuracy'] = self.class_wise_accuracy(event_label)
            results[event_label]['error_rate'] = self.class_wise_error_rate(event_label)
            results[event_label]['count'] = self.class_wise_count(event_label)

        return results

    def results_class_wise_average_metrics(self):
        """Class-wise averaged metrics

        Returns
        -------
        dict
            results in a dictionary format

        """

        event_wise_results = self.results_class_wise_metrics()

        event_wise_f_measure = []
        event_wise_precision = []
        event_wise_recall = []

        event_wise_error_rate = []
        event_wise_deletion_rate = []
        event_wise_insertion_rate = []

        event_wise_sensitivity = []
        event_wise_specificity = []
        event_wise_balanced_accuracy = []
        event_wise_accuracy = []

        for event_label in event_wise_results:
            # F-measure
            event_wise_f_measure.append(event_wise_results[event_label]['f_measure']['f_measure'])
            event_wise_precision.append(event_wise_results[event_label]['f_measure']['precision'])
            event_wise_recall.append(event_wise_results[event_label]['f_measure']['recall'])

            # Error rate
            event_wise_error_rate.append(event_wise_results[event_label]['error_rate']['error_rate'])
            event_wise_deletion_rate.append(event_wise_results[event_label]['error_rate']['deletion_rate'])
            event_wise_insertion_rate.append(event_wise_results[event_label]['error_rate']['insertion_rate'])

            # Accuracy
            if 'sensitivity' in event_wise_results[event_label]['accuracy']:
                event_wise_sensitivity.append(event_wise_results[event_label]['accuracy']['sensitivity'])

            if 'specificity' in event_wise_results[event_label]['accuracy']:
                event_wise_specificity.append(event_wise_results[event_label]['accuracy']['specificity'])

            if 'balanced_accuracy' in event_wise_results[event_label]['accuracy']:
                event_wise_balanced_accuracy.append(event_wise_results[event_label]['accuracy']['balanced_accuracy'])

            if 'accuracy' in event_wise_results[event_label]['accuracy']:
                event_wise_accuracy.append(event_wise_results[event_label]['accuracy']['accuracy'])

        if event_wise_f_measure:
            event_wise_f_measure_dict = {
                'f_measure': float(numpy.nanmean(event_wise_f_measure)),
                'precision': float(numpy.nanmean(event_wise_precision)),
                'recall': float(numpy.nanmean(event_wise_recall))
            }

        else:
            event_wise_f_measure_dict = {}

        if event_wise_error_rate:
            event_wise_error_rate_dict = {
                'error_rate': float(numpy.nanmean(event_wise_error_rate)),
                'deletion_rate': float(numpy.nanmean(event_wise_deletion_rate)),
                'insertion_rate': float(numpy.nanmean(event_wise_insertion_rate))
            }

        else:
            event_wise_error_rate_dict = {}

        if event_wise_accuracy:
            event_wise_accuracy_dict = {
                'sensitivity': float(numpy.nanmean(event_wise_sensitivity)),
                'specificity': float(numpy.nanmean(event_wise_specificity)),
                'balanced_accuracy': float(numpy.nanmean(event_wise_balanced_accuracy)),
                'accuracy': float(numpy.nanmean(event_wise_accuracy))
            }

        else:
            event_wise_accuracy_dict = {}

        return {
            'f_measure': event_wise_f_measure_dict,
            'error_rate': event_wise_error_rate_dict,
            'accuracy': event_wise_accuracy_dict
        }

    def results(self):
        """All metrics

        Returns
        -------
        dict
            results in a dictionary format

        """

        return {
            'overall': self.results_overall_metrics(),
            'class_wise': self.results_class_wise_metrics(),
            'class_wise_average': self.results_class_wise_average_metrics()
        }


class SegmentBasedMetrics(SoundEventMetrics):
    def __init__(self,
                 event_label_list,
                 time_resolution=1.0):
        """Constructor

        Parameters
        ----------
        event_label_list : list, numpy.array
            List of unique event labels

        time_resolution : float (0,]
            Segment size used in the evaluation, in seconds.
            Default value 1.0

        """

        SoundEventMetrics.__init__(self)

        if isinstance(event_label_list, numpy.ndarray) and len(event_label_list.shape) == 1:
            # We have numpy array, convert it to list
            event_label_list = event_label_list.tolist()

        if not isinstance(event_label_list, list):
            raise ValueError(
                "event_label_list needs to be list or numpy.array"
            )

        if not isinstance(time_resolution, float) or time_resolution <= 0.0:
            raise ValueError(
                "time_resolution needs to be float > 0"
            )

        self.event_label_list = event_label_list
        self.evaluated_length_seconds = 0.0
        self.evaluated_files = 0

        self.time_resolution = time_resolution

        self.overall = {
            'Ntp': 0.0,
            'Ntn': 0.0,
            'Nfp': 0.0,
            'Nfn': 0.0,
            'Nref': 0.0,
            'Nsys': 0.0,
            'ER': 0.0,
            'S': 0.0,
            'D': 0.0,
            'I': 0.0,
        }

        self.class_wise = {}
        for class_label in self.event_label_list:
            self.class_wise[class_label] = {
                'Ntp': 0.0,
                'Ntn': 0.0,
                'Nfp': 0.0,
                'Nfn': 0.0,
                'Nref': 0.0,
                'Nsys': 0.0,
            }

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        return self.results()

    def __str__(self):
        """Print result reports"""

        output = self.ui.section_header('Segment based metrics') + '\n'
        output += self.result_report_parameters() + '\n'
        output += self.result_report_overall() + '\n'
        output += self.result_report_class_wise_average() + '\n'
        output += self.result_report_class_wise() + '\n'

        return output

    def evaluate(self, reference_event_list, estimated_event_list, evaluated_length_seconds=None):
        """Evaluate file pair (reference and estimated)

        Parameters
        ----------

        reference_event_list : list of dict or dcase_util.containers.MetaDataContainer
            Reference event list.

        estimated_event_list : list of dict or dcase_util.containers.MetaDataContainer
            Estimated event list.

        evaluated_length_seconds : float, optional
            Evaluated length. If none given, maximum offset is used.
            Default value None

        Returns
        -------
        self

        """

        # Make sure input is dcase_util.containers.MetaDataContainer
        if not isinstance(reference_event_list, dcase_util.containers.MetaDataContainer):
            reference_event_list = dcase_util.containers.MetaDataContainer(reference_event_list)

        if not isinstance(estimated_event_list, dcase_util.containers.MetaDataContainer):
            estimated_event_list = dcase_util.containers.MetaDataContainer(estimated_event_list)

        # Check that input event list have event only from one file
        reference_files = reference_event_list.unique_files
        if len(reference_files) > 1:
            raise ValueError(
                "reference_event_list contains events from multiple files. Evaluate only file by file."
            )

        estimated_files = estimated_event_list.unique_files
        if len(estimated_files) > 1:
            raise ValueError(
                "estimated_event_list contains events from multiple files. Evaluate only file by file."
            )

        # Evaluate only valid events
        valid_reference_event_list = dcase_util.containers.MetaDataContainer()
        for item in reference_event_list:
            if 'event_onset' in item and 'event_offset' in item and 'event_label' in item:
                valid_reference_event_list.append(item)

            elif 'onset' in item and 'offset' in item and 'event_label' in item:
                valid_reference_event_list.append(item)

        reference_event_list = valid_reference_event_list

        valid_estimated_event_list = dcase_util.containers.MetaDataContainer()
        for item in estimated_event_list:
            if 'event_onset' in item and 'event_offset' in item and 'event_label' in item:
                valid_estimated_event_list.append(item)

            elif 'onset' in item and 'offset' in item and 'event_label' in item:
                valid_estimated_event_list.append(item)

        estimated_event_list = valid_estimated_event_list

        # Convert event list into frame-based representation
        reference_event_roll = util.event_list_to_event_roll(
            source_event_list=reference_event_list,
            event_label_list=self.event_label_list,
            time_resolution=self.time_resolution
        )

        estimated_event_roll = util.event_list_to_event_roll(
            source_event_list=estimated_event_list,
            event_label_list=self.event_label_list,
            time_resolution=self.time_resolution
        )

        if evaluated_length_seconds is None:
            evaluated_length_seconds = max(reference_event_list.max_offset, estimated_event_list.max_offset)
            evaluated_length_segments = int(math.ceil(evaluated_length_seconds * 1 / float(self.time_resolution)))

        else:
            evaluated_length_segments = int(math.ceil(evaluated_length_seconds * 1 / float(self.time_resolution)))

        self.evaluated_length_seconds += evaluated_length_seconds
        self.evaluated_files += 1

        reference_event_roll, estimated_event_roll = util.match_event_roll_lengths(
            reference_event_roll,
            estimated_event_roll,
            evaluated_length_segments
        )

        # Compute segment-based overall metrics
        for segment_id in range(0, reference_event_roll.shape[0]):
            annotated_segment = reference_event_roll[segment_id, :]
            system_segment = estimated_event_roll[segment_id, :]

            Ntp = sum(system_segment + annotated_segment > 1)
            Ntn = sum(system_segment + annotated_segment == 0)
            Nfp = sum(system_segment - annotated_segment > 0)
            Nfn = sum(annotated_segment - system_segment > 0)

            Nref = sum(annotated_segment)
            Nsys = sum(system_segment)

            S = min(Nref, Nsys) - Ntp
            D = max(0, Nref - Nsys)
            I = max(0, Nsys - Nref)

            self.overall['Ntp'] += Ntp
            self.overall['Ntn'] += Ntn
            self.overall['Nfp'] += Nfp
            self.overall['Nfn'] += Nfn
            self.overall['Nref'] += Nref
            self.overall['Nsys'] += Nsys
            self.overall['S'] += S
            self.overall['D'] += D
            self.overall['I'] += I

        # Compute segment-based class-wise metrics
        for class_id, class_label in enumerate(self.event_label_list):
            annotated_segment = reference_event_roll[:, class_id]
            system_segment = estimated_event_roll[:, class_id]

            Ntp = sum(system_segment + annotated_segment > 1)
            Ntn = sum(system_segment + annotated_segment == 0)
            Nfp = sum(system_segment - annotated_segment > 0)
            Nfn = sum(annotated_segment - system_segment > 0)

            Nref = sum(annotated_segment)
            Nsys = sum(system_segment)

            self.class_wise[class_label]['Ntp'] += Ntp
            self.class_wise[class_label]['Ntn'] += Ntn
            self.class_wise[class_label]['Nfp'] += Nfp
            self.class_wise[class_label]['Nfn'] += Nfn
            self.class_wise[class_label]['Nref'] += Nref
            self.class_wise[class_label]['Nsys'] += Nsys

        return self

    def reset(self):
        """Reset internal state"""

        self.overall = {
            'Ntp': 0.0,
            'Ntn': 0.0,
            'Nfp': 0.0,
            'Nfn': 0.0,
            'Nref': 0.0,
            'Nsys': 0.0,
            'ER': 0.0,
            'S': 0.0,
            'D': 0.0,
            'I': 0.0,
        }

        self.class_wise = {}
        for class_label in self.event_label_list:
            self.class_wise[class_label] = {
                'Ntp': 0.0,
                'Ntn': 0.0,
                'Nfp': 0.0,
                'Nfn': 0.0,
                'Nref': 0.0,
                'Nsys': 0.0,
            }

        return self

    # Metrics
    def overall_f_measure(self):
        """Overall f-measure metrics (f_measure, precision, and recall)

        Returns
        -------
        dict
            results in a dictionary format

        """
        if self.overall['Nsys'] == 0 and self.empty_system_output_handling == 'zero_score':
            precision = 0

        else:
            precision = metric.precision(
                Ntp=self.overall['Ntp'],
                Nsys=self.overall['Nsys']
            )

        recall = metric.recall(
            Ntp=self.overall['Ntp'],
            Nref=self.overall['Nref']
        )

        f_measure = metric.f_measure(
            precision=precision,
            recall=recall
        )

        return {
            'f_measure': f_measure,
            'precision': precision,
            'recall': recall
        }

    def overall_error_rate(self):
        """Overall error rate metrics (error_rate, substitution_rate, deletion_rate, and insertion_rate)

        Returns
        -------
        dict
            results in a dictionary format

        """

        substitution_rate = metric.substitution_rate(
            Nref=self.overall['Nref'],
            Nsubstitutions=self.overall['S']
        )

        deletion_rate = metric.deletion_rate(
            Nref=self.overall['Nref'],
            Ndeletions=self.overall['D']
        )

        insertion_rate = metric.insertion_rate(
            Nref=self.overall['Nref'],
            Ninsertions=self.overall['I']
        )

        error_rate = metric.error_rate(
            substitution_rate_value=substitution_rate,
            deletion_rate_value=deletion_rate,
            insertion_rate_value=insertion_rate
        )

        return {
            'error_rate': error_rate,
            'substitution_rate': substitution_rate,
            'deletion_rate': deletion_rate,
            'insertion_rate': insertion_rate
        }

    def overall_accuracy(self, factor=0.5):
        """Overall accuracy metrics (sensitivity, specificity, accuracy, and balanced_accuracy)

        Parameters
        ----------
        factor : float [0-1]
            Balance factor.
            Default value 0.5

        Returns
        -------
        dict
            results in a dictionary format

        """

        sensitivity = metric.sensitivity(
            Ntp=self.overall['Ntp'],
            Nfn=self.overall['Nfn']
        )

        specificity = metric.specificity(
            Ntn=self.overall['Ntn'],
            Nfp=self.overall['Nfp']
        )

        balanced_accuracy = metric.balanced_accuracy(
            sensitivity=sensitivity,
            specificity=specificity,
            factor=factor
        )

        accuracy = metric.accuracy(
            Ntp=self.overall['Ntp'],
            Ntn=self.overall['Ntn'],
            Nfp=self.overall['Nfp'],
            Nfn=self.overall['Nfn']
        )

        return {
            'accuracy': accuracy,
            'balanced_accuracy': balanced_accuracy,
            'sensitivity': sensitivity,
            'specificity': specificity
        }

    def class_wise_count(self, event_label):
        """Class-wise counts (Nref and Nsys)

        Returns
        -------
        dict
            results in a dictionary format

        """

        return {
            'Nref': float(self.class_wise[event_label]['Nref']),
            'Nsys': float(self.class_wise[event_label]['Nsys'])
        }

    def class_wise_f_measure(self, event_label):
        """Class-wise f-measure metrics (f_measure, precision, and recall)

        Returns
        -------
        dict
            results in a dictionary format

        """
        if self.class_wise[event_label]['Nsys'] == 0 and self.empty_system_output_handling == 'zero_score':
            precision = 0

        else:
            precision = metric.precision(
                Ntp=self.class_wise[event_label]['Ntp'],
                Nsys=self.class_wise[event_label]['Nsys']
            )

        recall = metric.recall(
            Ntp=self.class_wise[event_label]['Ntp'],
            Nref=self.class_wise[event_label]['Nref']
        )

        f_measure = metric.f_measure(
            precision=precision,
            recall=recall
        )

        return {
            'f_measure': f_measure,
            'precision': precision,
            'recall': recall
        }

    def class_wise_error_rate(self, event_label):
        """Class-wise error rate metrics (error_rate, deletion_rate, and insertion_rate)

        Returns
        -------
        dict
            results in a dictionary format

        """

        deletion_rate = metric.deletion_rate(
            Nref=self.class_wise[event_label]['Nref'],
            Ndeletions=self.class_wise[event_label]['Nfn']
        )

        insertion_rate = metric.insertion_rate(
            Nref=self.class_wise[event_label]['Nref'],
            Ninsertions=self.class_wise[event_label]['Nfp']
        )

        error_rate = metric.error_rate(
            deletion_rate_value=deletion_rate,
            insertion_rate_value=insertion_rate
        )

        return {
            'error_rate': error_rate,
            'deletion_rate': deletion_rate,
            'insertion_rate': insertion_rate
        }

    def class_wise_accuracy(self, event_label, factor=0.5):
        """Class-wise accuracy metrics (sensitivity, specificity, accuracy, and balanced_accuracy)

        Returns
        -------
        dict
            results in a dictionary format

        """

        sensitivity = metric.sensitivity(
            Ntp=self.class_wise[event_label]['Ntp'],
            Nfn=self.class_wise[event_label]['Nfn']
        )

        specificity = metric.specificity(
            Ntn=self.class_wise[event_label]['Ntn'],
            Nfp=self.class_wise[event_label]['Nfp']
        )

        balanced_accuracy = metric.balanced_accuracy(
            sensitivity=sensitivity,
            specificity=specificity,
            factor=factor
        )

        accuracy = metric.accuracy(
            Ntp=self.class_wise[event_label]['Ntp'],
            Ntn=self.class_wise[event_label]['Ntn'],
            Nfp=self.class_wise[event_label]['Nfp'],
            Nfn=self.class_wise[event_label]['Nfn']
        )

        return {
            'accuracy': accuracy,
            'balanced_accuracy': balanced_accuracy,
            'sensitivity': sensitivity,
            'specificity': specificity
        }

    # Reports
    def result_report_parameters(self):
        """Report metric parameters

        Returns
        -------
        str
            result report in string format

        """

        output = self.ui.data(field='Evaluated length', value=self.evaluated_length_seconds, unit='sec') + '\n'
        output += self.ui.data(field='Evaluated files', value=self.evaluated_files) + '\n'

        if self.time_resolution < 1:
            output += self.ui.data(field='Segment length', value=self.time_resolution * 1000, unit='ms') + '\n'

        else:
            output += self.ui.data(field='Segment length', value=self.time_resolution, unit='sec') + '\n'

        return output


class EventBasedMetrics(SoundEventMetrics):
    def __init__(self,
                 event_label_list,
                 evaluate_onset=True,
                 evaluate_offset=True,
                 onset_offset_op='and',
                 t_collar=0.200,
                 percentage_of_length=0.5,
                 event_matching_type='optimal',
                 **kwargs):
        """Constructor

        Parameters
        ----------
        event_label_list : list
            List of unique event labels

        evaluate_onset : bool
            Evaluate onset.
            Default value True

        evaluate_offset : bool
            Evaluate offset.
            Default value True

        t_collar : float (0,]
            Time collar used when evaluating validity of the onset and offset, in seconds.
            Default value 0.2

        percentage_of_length : float in [0, 1]
            Second condition, percentage of the length within which the estimated offset has to be in order to be
            consider valid estimation.
            Default value 0.5

        event_matching_type : str
            Event matching type. Set 'optimal' for graph-based matching, or 'greedy' for always select first found match.
            Greedy type of event matching is kept for backward compatibility. Both event matching types produce
            very similar results, however, greedy matching can be sensitive to the order of reference events.
            Use default 'optimal' event matching, if you do not intend to compare your results to old results.
            Default value 'optimal'

        """

        SoundEventMetrics.__init__(self, **kwargs)

        if isinstance(event_label_list, numpy.ndarray) and len(event_label_list.shape) == 1:
            # We have numpy array, convert it to list
            event_label_list = event_label_list.tolist()

        if not isinstance(event_label_list, list):
            raise ValueError(
                "event_label_list needs to be list or numpy.array"
            )

        if not isinstance(t_collar, float) or t_collar <= 0.0:
            raise ValueError(
                "t_collar needs to be float > 0"
            )

        if not isinstance(percentage_of_length, float) or percentage_of_length < 0.0 or percentage_of_length > 1.0:
            raise ValueError(
                "t_collar percentage_of_length to be float in [0, 1]"
            )

        self.event_label_list = event_label_list
        self.evaluated_length = 0.0
        self.evaluated_files = 0
        if not evaluate_onset and not evaluate_offset:
            raise ValueError("Both evaluate_onset and evaluate_offset cannot be set to False")

        self.evaluate_onset = evaluate_onset
        self.evaluate_offset = evaluate_offset

        # cxs mdfy
        self.onset_offset_op = onset_offset_op

        self.t_collar = t_collar
        self.percentage_of_length = percentage_of_length
        self.event_matching_type = event_matching_type
        self.overall = {
            'Nref': 0.0,
            'Nsys': 0.0,
            'Nsubs': 0.0,
            'Ntp': 0.0,
            'Nfp': 0.0,
            'Nfn': 0.0,
        }
        self.class_wise = {}

        for class_label in self.event_label_list:
            self.class_wise[class_label] = {
                'Nref': 0.0,
                'Nsys': 0.0,
                'Ntp': 0.0,
                'Ntn': 0.0,
                'Nfp': 0.0,
                'Nfn': 0.0,
            }

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        return self.results()

    def __str__(self):
        """Print result reports"""

        if self.evaluate_onset and self.evaluate_offset:
            title = "Event based metrics (onset-offset)"

        elif self.evaluate_onset and not self.evaluate_offset:
            title = "Event based metrics (onset only)"

        elif not self.evaluate_onset and self.evaluate_offset:
            title = "Event based metrics (offset only)"

        else:
            title = "Event based metrics"

        output = self.ui.section_header(title) + '\n'

        output += self.result_report_parameters() + '\n'
        output += self.result_report_overall() + '\n'
        output += self.result_report_class_wise_average() + '\n'
        output += self.result_report_class_wise() + '\n'

        return output

    def evaluate(self, reference_event_list, estimated_event_list):
        """Evaluate file pair (reference and estimated)

        Parameters
        ----------

        reference_event_list : event list
            Reference event list

        estimated_event_list : event list
            Estimated event list

        Returns
        -------
        self

        """

        # Make sure input is dcase_util.containers.MetaDataContainer
        if not isinstance(reference_event_list, dcase_util.containers.MetaDataContainer):
            reference_event_list = dcase_util.containers.MetaDataContainer(reference_event_list)

        if not isinstance(estimated_event_list, dcase_util.containers.MetaDataContainer):
            estimated_event_list = dcase_util.containers.MetaDataContainer(estimated_event_list)

        # Check that input event list have event only from one file
        reference_files = reference_event_list.unique_files
        if len(reference_files) > 1:
            raise ValueError(
                "reference_event_list contains events from multiple files. Evaluate only file by file."
            )

        estimated_files = estimated_event_list.unique_files
        if len(estimated_files) > 1:
            raise ValueError(
                "estimated_event_list contains events from multiple files. Evaluate only file by file."
            )

        # Evaluate only valid events
        valid_reference_event_list = dcase_util.containers.MetaDataContainer()
        for item in reference_event_list:
            if 'event_onset' in item and 'event_offset' in item and 'event_label' in item:
                valid_reference_event_list.append(item)

            elif 'onset' in item and 'offset' in item and 'event_label' in item:
                valid_reference_event_list.append(item)

        reference_event_list = valid_reference_event_list

        valid_estimated_event_list = dcase_util.containers.MetaDataContainer()
        for item in estimated_event_list:
            if 'event_onset' in item and 'event_offset' in item and 'event_label' in item:
                valid_estimated_event_list.append(item)

            elif 'onset' in item and 'offset' in item and 'event_label' in item:
                valid_estimated_event_list.append(item)

        estimated_event_list = valid_estimated_event_list

        self.evaluated_length += reference_event_list.max_offset
        self.evaluated_files += 1

        # Overall metrics

        # Total number of detected and reference events
        Nsys = len(estimated_event_list)
        Nref = len(reference_event_list)

        if self.event_matching_type == 'optimal':
            label_hit_matrix = numpy.zeros((len(reference_event_list), len(estimated_event_list)), dtype=bool)
            for j in range(0, len(reference_event_list)):
                for i in range(0, len(estimated_event_list)):
                    label_hit_matrix[j, i] = reference_event_list[j]['event_label'] == estimated_event_list[i][
                        'event_label']

            hit_matrix = label_hit_matrix
            if self.evaluate_onset:
                onset_hit_matrix = numpy.zeros((len(reference_event_list), len(estimated_event_list)), dtype=bool)
                for j in range(0, len(reference_event_list)):
                    for i in range(0, len(estimated_event_list)):
                        onset_hit_matrix[j, i] = self.validate_onset(
                            reference_event=reference_event_list[j],
                            estimated_event=estimated_event_list[i],
                            t_collar=self.t_collar
                        )

                hit_matrix *= onset_hit_matrix

            if self.evaluate_offset:
                offset_hit_matrix = numpy.zeros((len(reference_event_list), len(estimated_event_list)), dtype=bool)
                for j in range(0, len(reference_event_list)):
                    for i in range(0, len(estimated_event_list)):
                        offset_hit_matrix[j, i] = self.validate_offset(
                            reference_event=reference_event_list[j],
                            estimated_event=estimated_event_list[i],
                            t_collar=self.t_collar,
                            percentage_of_length=self.percentage_of_length
                        )
                if self.onset_offset_op == 'and':
                    hit_matrix *= offset_hit_matrix
                if self.onset_offset_op == 'or':
                    hit_matrix |= offset_hit_matrix

            hits = numpy.where(hit_matrix)
            G = {}
            for ref_i, est_i in zip(*hits):
                if est_i not in G:
                    G[est_i] = []

                G[est_i].append(ref_i)

            matching = sorted(util.bipartite_match(G).items())

            ref_correct = numpy.zeros(Nref, dtype=bool)
            sys_correct = numpy.zeros(Nsys, dtype=bool)
            for item in matching:
                ref_correct[item[0]] = True
                sys_correct[item[1]] = True

            Ntp = len(matching)
            # Substitutions
            Nsubs = 0

            ref_leftover = numpy.nonzero(numpy.logical_not(ref_correct))[0]
            sys_leftover = numpy.nonzero(numpy.logical_not(sys_correct))[0]
            sys_counted = numpy.zeros(Nsys, dtype=bool)
            for j in ref_leftover:
                for i in sys_leftover:
                    if not sys_counted[i]:
                        if self.evaluate_onset:
                            onset_condition = self.validate_onset(
                                reference_event=reference_event_list[j],
                                estimated_event=estimated_event_list[i],
                                t_collar=self.t_collar
                            )

                        else:
                            onset_condition = True

                        if self.evaluate_offset:
                            offset_condition = self.validate_offset(
                                reference_event=reference_event_list[j],
                                estimated_event=estimated_event_list[i],
                                t_collar=self.t_collar,
                                percentage_of_length=self.percentage_of_length
                            )

                        else:
                            offset_condition = True

                        if eval(f"{onset_condition} {self.onset_offset_op} {offset_condition}"):
                            sys_counted[i] = True
                            Nsubs += 1
                            break

        elif self.event_matching_type == 'greedy':
            sys_correct = numpy.zeros(Nsys, dtype=bool)
            ref_correct = numpy.zeros(Nref, dtype=bool)

            # Number of correctly detected events
            for j in range(0, len(reference_event_list)):
                for i in range(0, len(estimated_event_list)):
                    if not sys_correct[i]:  # skip already matched events
                        label_condition = reference_event_list[j]['event_label'] == estimated_event_list[i][
                            'event_label']

                        if self.evaluate_onset:
                            onset_condition = self.validate_onset(
                                reference_event=reference_event_list[j],
                                estimated_event=estimated_event_list[i],
                                t_collar=self.t_collar
                            )

                        else:
                            onset_condition = True

                        if self.evaluate_offset:
                            offset_condition = self.validate_offset(
                                reference_event=reference_event_list[j],
                                estimated_event=estimated_event_list[i],
                                t_collar=self.t_collar,
                                percentage_of_length=self.percentage_of_length
                            )

                        else:
                            offset_condition = True

                        if label_condition and (eval(f"{onset_condition} {self.onset_offset_op} {offset_condition}")):
                            ref_correct[j] = True
                            sys_correct[i] = True
                            break

            Ntp = numpy.sum(sys_correct)

            ref_leftover = numpy.nonzero(numpy.logical_not(ref_correct))[0]
            sys_leftover = numpy.nonzero(numpy.logical_not(sys_correct))[0]

            # Substitutions
            Nsubs = 0
            sys_counted = numpy.zeros(Nsys, dtype=bool)
            for j in ref_leftover:
                for i in sys_leftover:
                    if not sys_counted[i]:
                        if self.evaluate_onset:
                            onset_condition = self.validate_onset(
                                reference_event=reference_event_list[j],
                                estimated_event=estimated_event_list[i],
                                t_collar=self.t_collar
                            )

                        else:
                            onset_condition = True

                        if self.evaluate_offset:
                            offset_condition = self.validate_offset(
                                reference_event=reference_event_list[j],
                                estimated_event=estimated_event_list[i],
                                t_collar=self.t_collar,
                                percentage_of_length=self.percentage_of_length
                            )

                        else:
                            offset_condition = True

                        if eval(f"{onset_condition} {self.onset_offset_op} {offset_condition}"):
                            sys_counted[i] = True
                            Nsubs += 1
                            break

        Nfp = Nsys - Ntp - Nsubs
        Nfn = Nref - Ntp - Nsubs

        self.overall['Nref'] += Nref
        self.overall['Nsys'] += Nsys
        self.overall['Ntp'] += Ntp
        self.overall['Nsubs'] += Nsubs
        self.overall['Nfp'] += Nfp
        self.overall['Nfn'] += Nfn

        # Class-wise metrics
        for class_id, class_label in enumerate(self.event_label_list):
            Nref = 0.0
            Nsys = 0.0
            Ntp = 0.0

            # Count event frequencies in the ground truth
            for i in range(0, len(reference_event_list)):
                if reference_event_list[i]['event_label'] == class_label:
                    Nref += 1

            # Count event frequencies in the system output
            for i in range(0, len(estimated_event_list)):
                if estimated_event_list[i]['event_label'] == class_label:
                    Nsys += 1

            if self.event_matching_type == 'optimal':
                class_reference_event_list = reference_event_list.filter(event_label=class_label)
                class_estimated_event_list = estimated_event_list.filter(event_label=class_label)

                hit_matrix = numpy.ones((len(class_reference_event_list), len(class_estimated_event_list)), dtype=bool)
                if self.evaluate_onset:
                    onset_hit_matrix = numpy.zeros((len(class_reference_event_list), len(class_estimated_event_list)),
                                                   dtype=bool)
                    for j in range(0, len(class_reference_event_list)):
                        for i in range(0, len(class_estimated_event_list)):
                            onset_hit_matrix[j, i] = self.validate_onset(
                                reference_event=class_reference_event_list[j],
                                estimated_event=class_estimated_event_list[i],
                                t_collar=self.t_collar
                            )

                    hit_matrix *= onset_hit_matrix

                if self.evaluate_offset:
                    offset_hit_matrix = numpy.zeros((len(class_reference_event_list), len(class_estimated_event_list)),
                                                    dtype=bool)
                    for j in range(0, len(class_reference_event_list)):
                        for i in range(0, len(class_estimated_event_list)):
                            offset_hit_matrix[j, i] = self.validate_offset(
                                reference_event=class_reference_event_list[j],
                                estimated_event=class_estimated_event_list[i],
                                t_collar=self.t_collar,
                                percentage_of_length=self.percentage_of_length
                            )
                    if self.onset_offset_op == 'and':
                        hit_matrix *= offset_hit_matrix
                    if self.onset_offset_op == 'or':
                        hit_matrix |= offset_hit_matrix

                hits = numpy.where(hit_matrix)
                G = {}
                for ref_i, est_i in zip(*hits):
                    if est_i not in G:
                        G[est_i] = []

                    G[est_i].append(ref_i)

                matching = sorted(util.bipartite_match(G).items())

                ref_correct = numpy.zeros(int(Nref), dtype=bool)
                sys_correct = numpy.zeros(int(Nsys), dtype=bool)

                for item in matching:
                    ref_correct[item[0]] = True
                    sys_correct[item[1]] = True

                Ntp = len(matching)

            elif self.event_matching_type == 'greedy':
                sys_counted = numpy.zeros(len(estimated_event_list), dtype=bool)
                for j in range(0, len(reference_event_list)):
                    if reference_event_list[j]['event_label'] == class_label:
                        for i in range(0, len(estimated_event_list)):
                            if estimated_event_list[i]['event_label'] == class_label and not sys_counted[i]:
                                if self.evaluate_onset:
                                    onset_condition = self.validate_onset(
                                        reference_event=reference_event_list[j],
                                        estimated_event=estimated_event_list[i],
                                        t_collar=self.t_collar
                                    )

                                else:
                                    onset_condition = True

                                if self.evaluate_offset:
                                    offset_condition = self.validate_offset(
                                        reference_event=reference_event_list[j],
                                        estimated_event=estimated_event_list[i],
                                        t_collar=self.t_collar,
                                        percentage_of_length=self.percentage_of_length
                                    )

                                else:
                                    offset_condition = True

                                if eval(f"{onset_condition} {self.onset_offset_op} {offset_condition}"):
                                    sys_counted[i] = True
                                    Ntp += 1
                                    break

            Nfp = Nsys - Ntp
            Nfn = Nref - Ntp

            self.class_wise[class_label]['Nref'] += Nref
            self.class_wise[class_label]['Nsys'] += Nsys

            self.class_wise[class_label]['Ntp'] += Ntp
            self.class_wise[class_label]['Nfp'] += Nfp
            self.class_wise[class_label]['Nfn'] += Nfn

            # modfy by cxs 
            self.class_wise[class_label]['sysCorrectEvent'] = sys_correct
            self.class_wise[class_label]['refCorrectEvent'] = ref_correct
            # end

        return self

    def reset(self):
        """Reset internal state
        """

        self.overall = {
            'Nref': 0.0,
            'Nsys': 0.0,
            'Nsubs': 0.0,
            'Ntp': 0.0,
            'Nfp': 0.0,
            'Nfn': 0.0,
        }
        self.class_wise = {}

        for class_label in self.event_label_list:
            self.class_wise[class_label] = {
                'Nref': 0.0,
                'Nsys': 0.0,
                'Ntp': 0.0,
                'Ntn': 0.0,
                'Nfp': 0.0,
                'Nfn': 0.0,
            }

        return self

    @staticmethod
    def validate_onset(reference_event, estimated_event, t_collar=0.200):
        """Validate estimated event based on event onset

        Parameters
        ----------
        reference_event : dict
            Reference event.

        estimated_event: dict
            Estimated event.

        t_collar : float > 0, seconds
            Time collar with which the estimated onset has to be in order to be consider valid estimation.
            Default value 0.2

        Returns
        -------
        bool

        """

        # Detect field naming style used and validate onset
        if 'event_onset' in reference_event and 'event_onset' in estimated_event:
            return math.fabs(reference_event['event_onset'] - estimated_event['event_onset']) <= t_collar

        elif 'onset' in reference_event and 'onset' in estimated_event:
            return math.fabs(reference_event['onset'] - estimated_event['onset']) <= t_collar

    @staticmethod
    def validate_offset(reference_event, estimated_event, t_collar=0.200, percentage_of_length=0.5):
        """Validate estimated event based on event offset

        Parameters
        ----------
        reference_event : dict
            Reference event.

        estimated_event : dict
            Estimated event.

        t_collar : float > 0, seconds
            First condition, Time collar with which the estimated offset has to be in order to be consider valid estimation.
            Default value 0.2

        percentage_of_length : float in [0, 1]
            Second condition, percentage of the length within which the estimated offset has to be in order to be
            consider valid estimation.
            Default value 0.5

        Returns
        -------
        bool

        """

        # Detect field naming style used and validate onset
        if 'event_offset' in reference_event and 'event_offset' in estimated_event:
            annotated_length = reference_event['event_offset'] - reference_event['event_onset']

            return math.fabs(reference_event['event_offset'] - estimated_event['event_offset']) <= max(t_collar,
                                                                                                       percentage_of_length * annotated_length)

        elif 'offset' in reference_event and 'offset' in estimated_event:
            annotated_length = reference_event['offset'] - reference_event['onset']

            return math.fabs(reference_event['offset'] - estimated_event['offset']) <= max(t_collar,
                                                                                           percentage_of_length * annotated_length)

    # Metrics
    def overall_f_measure(self):
        """Overall f-measure metrics (f_measure, precision, and recall)

        Returns
        -------
        dict
            results in a dictionary format
        """

        if self.overall['Nsys'] == 0 and self.empty_system_output_handling == 'zero_score':
            precision = 0

        else:
            precision = metric.precision(
                Ntp=self.overall['Ntp'],
                Nsys=self.overall['Nsys']
            )

        recall = metric.recall(
            Ntp=self.overall['Ntp'],
            Nref=self.overall['Nref']
        )

        f_measure = metric.f_measure(
            precision=precision,
            recall=recall
        )

        return {
            'f_measure': f_measure,
            'precision': precision,
            'recall': recall
        }

    def overall_error_rate(self):
        """Overall error rate metrics (error_rate, substitution_rate, deletion_rate, and insertion_rate)

        Returns
        -------
        dict
            results in a dictionary format

        """

        substitution_rate = metric.substitution_rate(
            Nref=self.overall['Nref'],
            Nsubstitutions=self.overall['Nsubs']
        )

        deletion_rate = metric.deletion_rate(
            Nref=self.overall['Nref'],
            Ndeletions=self.overall['Nfn']
        )

        insertion_rate = metric.insertion_rate(
            Nref=self.overall['Nref'],
            Ninsertions=self.overall['Nfp']
        )

        error_rate = metric.error_rate(
            substitution_rate_value=substitution_rate,
            deletion_rate_value=deletion_rate,
            insertion_rate_value=insertion_rate
        )

        return {
            'error_rate': error_rate,
            'substitution_rate': substitution_rate,
            'deletion_rate': deletion_rate,
            'insertion_rate': insertion_rate
        }

    def class_wise_count(self, event_label):
        """Class-wise counts (Nref and Nsys)

        Returns
        -------
        dict
            results in a dictionary format

        """

        return {
            'Nref': self.class_wise[event_label]['Nref'],
            'Nsys': self.class_wise[event_label]['Nsys']
        }

    def class_wise_f_measure(self, event_label):
        """Class-wise f-measure metrics (f_measure, precision, and recall)

        Returns
        -------
        dict
            results in a dictionary format

        """
        if self.class_wise[event_label]['Nsys'] == 0 and self.empty_system_output_handling == 'zero_score':
            precision = 0

        else:
            precision = metric.precision(
                Ntp=self.class_wise[event_label]['Ntp'],
                Nsys=self.class_wise[event_label]['Nsys']
            )

        recall = metric.recall(
            Ntp=self.class_wise[event_label]['Ntp'],
            Nref=self.class_wise[event_label]['Nref']
        )

        f_measure = metric.f_measure(
            precision=precision,
            recall=recall
        )

        return {
            'f_measure': f_measure,
            'precision': precision,
            'recall': recall
        }

    def class_wise_error_rate(self, event_label):
        """Class-wise error rate metrics (error_rate, deletion_rate, and insertion_rate)

        Returns
        -------
        dict
            results in a dictionary format

        """

        deletion_rate = metric.deletion_rate(
            Nref=self.class_wise[event_label]['Nref'],
            Ndeletions=self.class_wise[event_label]['Nfn']
        )

        insertion_rate = metric.insertion_rate(
            Nref=self.class_wise[event_label]['Nref'],
            Ninsertions=self.class_wise[event_label]['Nfp']
        )

        error_rate = metric.error_rate(
            deletion_rate_value=deletion_rate,
            insertion_rate_value=insertion_rate
        )

        return {
            'error_rate': error_rate,
            'deletion_rate': deletion_rate,
            'insertion_rate': insertion_rate
        }

    # Reports
    def result_report_parameters(self):
        """Report metric parameters

        Returns
        -------
        str
            result report in string format

        """

        output = self.ui.data(field='Evaluated length', value=self.evaluated_length, unit='sec') + '\n'
        output += self.ui.data(field='Evaluated files', value=self.evaluated_files) + '\n'

        output += self.ui.data(field='Evaluate onset', value=self.evaluate_onset) + '\n'
        output += self.ui.data(field='Evaluate offset', value=self.evaluate_offset) + '\n'

        if self.t_collar < 1:
            output += self.ui.data(field='T collar', value=self.t_collar * 1000, unit='ms') + '\n'

        else:
            output += self.ui.data(field='T collar', value=self.t_collar, unit='sec') + '\n'

        output += self.ui.data(field='Offset (length)', value=self.percentage_of_length * 100, unit='%') + '\n'

        return output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

.. autosummary::
    :toctree: generated/

    f_measure
    precision
    recall
    
    accuracy
    accuracy2
    balanced_accuracy
    sensitivity
    specificity

    error_rate
    substitution_rate
    deletion_rate
    insertion_rate

    equal_error_rate

"""

import numpy


# -- F-measure -- #
def precision(Ntp, Nsys, eps=numpy.spacing(1)):
    """Precision.

    Wikipedia entry https://en.wikipedia.org/wiki/Precision_and_recall

    Parameters
    ----------
    Ntp : int >=0
        Number of true positives.

    Nsys : int >=0
        Amount of system output.

    eps : float
        eps.
        Default value numpy.spacing(1)

    Returns
    -------
    precision: float
        Precision

    """

    if Nsys == 0:
        return numpy.nan
    else:
        return float(Ntp / float(Nsys))


def recall(Ntp, Nref, eps=numpy.spacing(1)):
    """Recall.

    Wikipedia entry https://en.wikipedia.org/wiki/Precision_and_recall

    Parameters
    ----------
    Ntp : int >=0
        Number of true positives.

    Nref : int >=0
        Amount of reference.

    eps : float
        eps.
        Default value numpy.spacing(1)

    Returns
    -------
    recall: float
        Recall

    """

    if Nref == 0:
        return numpy.nan
    else:
        return float(Ntp / float(Nref))


def f_measure(precision, recall, beta=1.0):
    """F-measure from precision and recall scores.

    Wikipedia entry https://en.wikipedia.org/wiki/F1_score

    Parameters
    ----------
    precision : float in (0, 1]
        Precision.

    recall : float in (0, 1]
        Recall.

    beta : float > 0
        Weighting factor for f-measure.
        Default value 1.0

    Returns
    -------
    f_measure: float
        The weighted f-measure

    """

    if precision == 0 and recall == 0:
        return 0.0

    return float((1 + beta**2)*precision*recall/((beta**2)*precision + recall))


# -- Accuracy -- #
def sensitivity(Ntp, Nfn, eps=numpy.spacing(1)):
    """Sensitivity

    Wikipedia entry https://en.wikipedia.org/wiki/Sensitivity_and_specificity

    Parameters
    ----------
    Ntp : int >=0
        Number of true positives.

    Nfn : int >=0
        Number of false negatives.

    eps : float
        eps.
        Default value numpy.spacing(1)

    Returns
    -------
    sensitivity: float
        Sensitivity

    """

    return float(Ntp / (Ntp + Nfn + eps))


def specificity(Ntn, Nfp, eps=numpy.spacing(1)):
    """Specificity

    Wikipedia entry https://en.wikipedia.org/wiki/Sensitivity_and_specificity

    Parameters
    ----------
    Ntn : int >= 0
        Number of true negatives.

    Nfp : int >= 0
        Number of false positives.
    
    eps : float
        eps.
        Default value numpy.spacing(1)

    Returns
    -------
    specificity: float
        Specificity

    """

    return float(Ntn / (Ntn + Nfp + eps))


def balanced_accuracy(sensitivity, specificity, factor=0.5):
    """Balanced accuracy

    Wikipedia entry https://en.wikipedia.org/wiki/Accuracy_and_precision

    Parameters
    ----------
    sensitivity : float in [0, 1]
        sensitivity.

    specificity : float in [0, 1]
        specificity.

    factor : float in [0, 1]
        Balancing factor multiplying true positive rate (sensitivity).
        Default value 0.5

    Returns
    -------
    bacc: float
        Balanced accuracy

    """

    return float(((1-factor) * sensitivity) + (factor * specificity))


def accuracy(Ntp, Ntn, Nfp, Nfn, eps=numpy.spacing(1)):
    """Accuracy

    Parameters
    ----------
    Ntp : int >= 0
        Number of true positives.

    Ntn : int >= 0
        Number of true negatives.

    Nfp : int >= 0
        Number of false positives.

    Nfn : int >= 0
        Number of false negatives.

    eps : float
        eps.
        Default value numpy.spacing(1)

    Returns
    -------
    acc: float
        Accuracy

    """

    return float((Ntp + Ntn) / (Ntp + Ntn + Nfn + Nfp + eps))


def accuracy_corr(Ncorr, N, eps=numpy.spacing(1)):
    """Accuracy

    Parameters
    ----------
    Ncorr : int >= 0
        Number of correct items.

    N : int >= 0
        Number of all items.

    eps : float
        eps.
        Default value numpy.spacing(1)

    Returns
    -------
    acc: float
        Accuracy.

    """

    return float((Ncorr) / (N + eps))


def accuracy2(Ntp, Nfp, Nfn, eps=numpy.spacing(1)):
    """Accuracy

    Implements metric described by [1]_.

    .. [1] Dixon, S.
           "On the computer recognition of solo piano music".
           Proceedings of Australasian Computer Music Conference, July 2000, Brisbane, Australia, pp. 31–37.
           http://www.eecs.qmul.ac.uk/~simond/pub/2000/acmc.pdf

    Parameters
    ----------
    Ntp : int >= 0
        Number of true positives.

    Nfp : int >= 0
        Number of false positives.

    Nfn : int >= 0
        Number of false negatives.

    eps : float
        eps.
        Default value numpy.spacing(1)

    Returns
    -------
    acc: float
        Accuracy

    """

    return float((Ntp) / (Ntp + Nfn + Nfp + eps))


def substitution_rate(Nref, Nsubstitutions, eps=numpy.spacing(1)):
    """Substitution rate

    Parameters
    ----------
    Nref : int >=0
        Number of entries in the reference.

    Nsubstitutions : int >=0
        Number of substitutions.

    eps : float
        eps.
        Default value numpy.spacing(1)

    Returns
    -------
    substitution_rate: float
        Substitution rate

    """

    return float(Nsubstitutions / (Nref + eps))


def deletion_rate(Nref, Ndeletions, eps=numpy.spacing(1)):
    """Deletion rate

    Parameters
    ----------
    Nref : int >=0
        Number of entries in the reference.

    Ndeletions : int >=0
        Number of deletions.

    eps : float
        eps.
        Default value numpy.spacing(1)

    Returns
    -------
    deletion_rate: float
        Deletion rate
        
    """

    return float(Ndeletions / (Nref + eps))


def insertion_rate(Nref, Ninsertions, eps=numpy.spacing(1)):
    """Insertion rate

    Parameters
    ----------
    Nref : int >=0
        Number of entries in the reference.

    Ninsertions : int >=0
        Number of insertions.

    eps : float
        eps.
        Default value numpy.spacing(1)

    Returns
    -------
    insertion_rate: float
        Insertion rate

    """

    return float(Ninsertions / (Nref + eps))


def error_rate(substitution_rate_value=0.0, deletion_rate_value=0.0, insertion_rate_value=0.0):
    """Error rate

    Parameters
    ----------
    substitution_rate_value : float >=0
        Substitution rate.
        Default value 0

    deletion_rate_value : float >=0
        Deletion rate.
        Default value 0

    insertion_rate_value : float >=0
        Insertion rate.
        Default value 0

    Returns
    -------
    error_rate: float
        Error rate

    """

    return float(substitution_rate_value + deletion_rate_value + insertion_rate_value)


def equal_error_rate(y_true, y_score, eps=numpy.spacing(1)):
    """Equal error rate (EER)

    EER is calculated from the curve of the false negative rate versus the false positive rate.
    Implementation is based on https://github.com/pafoster/dcase2016_task4/blob/master/evaluation_scripts/eer.py

    Parameters
    ----------
    y_true : numpy.array or list
        True binary labels in range {0, 1} or {-1, 1}.

    y_score : numpy.array or list
        Target scores, can either be probability estimates of the positive
        class or confidence values.

    eps : float
        Minimum difference considered equal.
        Default value numpy.spacing(1)

    Returns
    -------
    float

    """

    from sklearn import metrics

    if numpy.any(y_true):
        false_positive_rate, true_positive_rate, thresholds = metrics.roc_curve(
            y_true=y_true,
            y_score=y_score,
            drop_intermediate=True
        )

        points = [(0, 0)] + list(zip(false_positive_rate, true_positive_rate))
        for i, point in enumerate(points):
            if point[0] + eps >= 1 - point[1]:
                break

        point1 = points[i - 1]
        point2 = points[i]

        # Interpolate between point1 and point2
        if abs(point2[0] - point1[0]) < eps:
            eer = point1[0]

        else:
            m = (point2[1] - point1[1]) / (point2[0] - point1[0])
            o = point1[1] - m * point1[0]
            eer = (1 - o) / (1 + m)

    else:
        eer = numpy.nan

    return eer



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   plot_parallel_data 
@Time        :   2023/11/30 18:32
@Author      :   Xuesong Chen
@Description :   
"""
import matplotlib.pyplot as plt

def plot_multiple_datasets(*datasets, sampling_rates, labels):
    """
    Plots multiple datasets with different sampling rates but the same total duration.

    :param datasets: A variable number of datasets (arrays or lists of data points).
    :param sampling_rates: A list of sampling rates corresponding to each dataset.
    :param labels: A list of labels corresponding to each dataset.
    """
    # 创建一个图形和多个轴
    fig, axes = plt.subplots(len(datasets), 1, sharex=True, figsize=(10, 6))

    # 如果只有一个数据集，我们需要将axes数组化
    if len(datasets) == 1:
        axes = [axes]

    # 为每个数据集创建一个轴
    for i, data in enumerate(datasets):
        time_axis = [t / sampling_rates[i] for t in range(len(data))]
        axes[i].plot(time_axis, data)
        axes[i].set_ylabel(labels[i])

    # 设置x轴标签
    plt.xlabel('Time (s)')

    # 显示图表
    plt.show()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   detect_oxygen_desaturation
@Time        :   2023/4/25 14:41
@Author      :   Xuesong Chen
@Description :
"""
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

SPO2_MIN_VALUE = 110

def detect_oxygen_desaturation(spo2_arr, duration_max=120, spo2_des_min_thre=3, ret_format='df'):
    '''

    :param spo2_arr:
    :param duration_max:
    :param spo2_des_min_thre: 血氧降低最低报警阈值
    :return:
    '''
    # assert np.min(spo2_arr) >= 0 and np.max(spo2_arr) <= 100, 'Invalid SpO2 value'
    spo2_max = spo2_arr[0]  # 初始化最大值
    spo2_max_index = 1  # 初始化最大值下标
    spo2_min = SPO2_MIN_VALUE  # 初始化血氧最低值
    des_onset_pred_set = np.array([], dtype=int)  # 算法预测氧降起始点总集
    des_duration_pred_set = np.array([], dtype=int)  # 算法预测氧降持续时间总集
    des_level_set = np.array([])  # 被记录的氧降事件合集 （下降百分比2%，3%，4%，5%....
    des_onset_pred_point = 0  # 预测事件起始点
    des_flag = 0  # 氧降事件发生标记
    ma_flag = 0  # motion artifact事件发生标记
    spo2_des_max_thre = 50  # 血氧motion artifact阈值,50%
    duration_min = 5  # 氧降事件至少持续duration_min s才会被记录在内
    prob_end = []

    for i, current_value in enumerate(spo2_arr):

        des_percent = spo2_max - current_value  # 氧降值

        # 检测Motion artifacts
        if ma_flag and (des_percent < spo2_des_max_thre):
            if des_flag and len(prob_end) != 0:
                des_onset_pred_set = np.append(des_onset_pred_set, des_onset_pred_point)
                des_duration_pred_set = np.append(des_duration_pred_set, prob_end[-1] - des_onset_pred_point)
                des_level_point = spo2_max - spo2_min
                des_level_set = np.append(des_level_set, des_level_point)
            # 重置
            spo2_max = current_value
            spo2_max_index = i
            ma_flag = 0
            des_flag = 0
            spo2_min = SPO2_MIN_VALUE
            prob_end = []
            continue

        # 如果氧降值大于2%, 记录起始时间
        if des_percent >= spo2_des_min_thre:
            # 如果氧降程度过大，或者
            # todo 血氧值突然下降，认为是motion artifact
            if des_percent > spo2_des_max_thre:
                ma_flag = 1
            else:
                des_onset_pred_point = spo2_max_index
                des_flag = 1
                if current_value < spo2_min:
                    spo2_min = current_value

        if current_value >= spo2_max and not des_flag:
            spo2_max = current_value
            spo2_max_index = i

        elif des_flag:

            if current_value > spo2_min:
                if current_value > spo2_arr[i - 1]:
                    prob_end.append(i)

                # 定位血氧连续下降点
                if current_value <= spo2_arr[i - 1] < spo2_arr[i - 2]:
                    spo2_des_duration = prob_end[-1] - spo2_max_index

                    # 1. 下降时间不够的话，则不认为是氧降事件
                    if spo2_des_duration < duration_min:
                        spo2_max = spo2_arr[i - 2]
                        spo2_max_index = i - 2
                        spo2_min = SPO2_MIN_VALUE
                        des_flag = 0
                        prob_end = []
                        continue

                    else:
                        # 下降时间满足条件，则记录该次氧降事件
                        if duration_min <= spo2_des_duration <= duration_max:
                            des_onset_pred_set = np.append(des_onset_pred_set, des_onset_pred_point)
                            des_duration_pred_set = np.append(des_duration_pred_set, spo2_des_duration)
                            des_level_point = spo2_max - spo2_min
                            des_level_set = np.append(des_level_set, des_level_point)

                        # 下降时间过长，说明存在多个氧降事件，需要分开记录
                        else:
                            # 记录第一个氧降事件
                            des_onset_pred_set = np.append(des_onset_pred_set, des_onset_pred_point)
                            des_duration_pred_set = np.append(des_duration_pred_set, prob_end[0] - des_onset_pred_point)
                            des_level_point = spo2_max - spo2_min
                            des_level_set = np.append(des_level_set, des_level_point)

                            # 重新查找可能的氧降事件
                            remain_spo2_arr = spo2_arr[prob_end[0]:i + 1]
                            _onset, _duration, _des_level = detect_oxygen_desaturation(remain_spo2_arr,
                                                                                       ret_format='tuple')
                            des_onset_pred_set = np.append(des_onset_pred_set, _onset + prob_end[0])
                            des_duration_pred_set = np.append(des_duration_pred_set, _duration)
                            des_level_set = np.append(des_level_set, _des_level)

                        spo2_max = spo2_arr[i - 2]
                        spo2_max_index = i - 2
                        spo2_min = SPO2_MIN_VALUE
                        des_flag = 0
                        prob_end = []
    if ret_format == 'tuple':
        return des_onset_pred_set, des_duration_pred_set, des_level_set
    else:
        return pd.DataFrame({
            'Type': 'OD',
            'Start': des_onset_pred_set,
            'Duration': des_duration_pred_set,
            'OD_level': des_level_set
        })

if __name__ == '__main__':
    data = np.load('/Users/cxs/Downloads/gt_spo2.npz')
    from wuji.tools.plot_labeled_signal import plot_signal_with_events
    gt_spo2 = data['spo2']
    res = detect_oxygen_desaturation(gt_spo2, ret_format='df')
    plot_signal_with_events(gt_spo2, 1, res)
    # plt.plot(gt_spo2)
    # plt.show()


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   utils 
@Time        :   2023/9/14 17:33
@Author      :   Xuesong Chen
@Description :   
"""

import pandas as pd


def get_equal_duration_and_labeled_chunks(event_df, chunk_duration=30):
    """将标注数据框转换为等长事件列表，如果最后一个事件的持续时长不足chunk_duration，则以真实值填充。

        Parameters
        ----------
        event_df : pandas.DataFrame
            具有标注名称(Type)、开始时长(Start)和持续时长(Duration)的数据框。

        Returns
        -------
        ret_df : pandas.DataFrame
            具有标注名称(Type)、开始时长(Start)和持续时长(Duration)的数据框。
        """
    assert 'Type' in event_df.columns and 'Start' in event_df.columns and 'Duration' in event_df.columns, "check columns"

    ret_df = pd.DataFrame(columns=['Type', 'Start', 'Duration'])
    for idx, row in event_df.iterrows():
        start = int(row['Start'])
        duration = row['Duration']
        while duration > 0:
            cur_duration = duration if duration < chunk_duration else chunk_duration
            ret_df = pd.concat([
                ret_df,
                pd.DataFrame({'Type': row['Type'], 'Start': start, 'Duration': cur_duration}, index=[0])])
            start += chunk_duration
            duration -= chunk_duration
    assert sum(ret_df['Duration'] % chunk_duration != 0) <= 1, "sleep stage duration error"
    return ret_df.reset_index(drop=True)


import pandas as pd


def get_irequal_duration_chunks(chunks_df):
    """将等长事件列表还原为原始格式。

    Parameters
    ----------
    chunks_df : pandas.DataFrame
        具有标注名称(Type)、开始时长(Start)和持续时长(Duration)的数据框。

    Returns
    -------
    original_df : pandas.DataFrame
        还原后的数据框。
    """
    # 按照 Type 分组，并在每组内按照 Start 排序

    grouped = chunks_df.groupby('Type').apply(lambda x: x.sort_values('Start')).reset_index(drop=True)

    merged_chunks = []
    for _, group in chunks_df.groupby('Type'):
        current_type = group.iloc[0]['Type']
        current_start = group.iloc[0]['Start']
        current_duration = 0

        for _, row in group.iterrows():
            # If the current row is a continuation of the chunk, increase the duration
            if row['Start'] == current_start + current_duration:
                current_duration += row['Duration']
            else:
                # If it's not a continuation, save the current chunk and start a new one
                if current_duration > 0:  # Avoid appending zero-duration chunks
                    merged_chunks.append({'Type': current_type, 'Start': current_start, 'Duration': current_duration})
                current_start = row['Start']
                current_duration = row['Duration']

        merged_chunks.append({'Type': current_type, 'Start': current_start, 'Duration': current_duration})

    return pd.DataFrame(merged_chunks).sort_values('Start').reset_index(drop=True)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   __init__.py 
@Time        :   2023/9/21 21:45
@Author      :   Xuesong Chen
@Description :   
"""

from .Classification import ClassificationEvaluator
from .EventDetector import EventDetectorEvaluator
from .NightlySleepMetrics import NightlySleepMetricsEvaluator